/* 
 * SNFS API Layer
 * 
 * snfs_api.c
 *
 * Implementation of the SNFS API layer of the client side SNFS library.
 * 
 * Read the SNFS API interface (snfs_api.h) and the project specification 
 * for futher information about the services.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <strings.h>

#include <snfs_api.h>
#include <snfs_proto.h>


#define dprintf if (1) printf


/*
 * Internal private global variables
 */

// the client socket file descriptor
static int Cli_sock;

// the client socket address
static struct sockaddr_un Cli_addr;

// the server socket address
static struct sockaddr_un Serv_addr;


/*
 * Internal private auxiliary functions
 */


static int remote_call(snfs_msg_req_t *req, int reqsz, snfs_msg_res_t *res, 
		       int ressz)
{
   int status;
   
   // sends the request to the server
   status = sendto(Cli_sock, (void*)req, reqsz, 0, 
      (struct sockaddr *)&Serv_addr, sizeof(Serv_addr));
   if (status < 0) {
      printf("[snfs_api] sendto error: %s.\n", strerror(errno));
      return -1;
   }
  
   // waits for an answer
   status = recvfrom(Cli_sock, res, ressz, 0, NULL, NULL);
   if (status < 0) {
      printf("[snfs_api] recvfrom error: %s.\n", strerror(errno));
      return -1;
   }
   if (status == 0) {
      printf("[snfs_api] server is closed.\n");
      return -1;
   }

   return status;
}


/*
 * SNFS API implementation (see snfs_api.h)
 */


int snfs_init(char* cli_name, char* server_name)
{
   if (cli_name == NULL || server_name == NULL) {
      printf("[snfs_api] invalid client/server address names.\n");
      return -1;
   }

   // creates socket datagram domain internet
   if ((Cli_sock = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0){
      printf("[snfs_api] socket error: %s.\n", strerror(errno));
      return -1;
   }

   // reuse client address if it exists
   if (unlink(cli_name) < 0 && errno != ENOENT) {
      printf("[snfs_api] unlink error: %s.\n", strerror(errno));
   }

   // client structure address cleaning
   bzero(&Cli_addr, sizeof(Cli_addr));
   Cli_addr.sun_family = AF_UNIX;
   strcpy(Cli_addr.sun_path, cli_name);

   // binds socket to address
   if (bind(Cli_sock, (struct sockaddr *) &Cli_addr, sizeof(Cli_addr)) < 0){
      printf("[snfs_api] bind error: %s.\n", strerror(errno));
      return -1;
   }

   // server structure address cleaning
   bzero(&Serv_addr, sizeof(Serv_addr));
   Serv_addr.sun_family = AF_UNIX;
   strcpy(Serv_addr.sun_path, server_name);

   return 0;
}


snfs_call_status_t snfs_ping(char* inmsg, int insize, char* outmsg,
			     int outsize)
{
  if (inmsg == NULL || insize < 0 || insize > 256 || outmsg == NULL
      || outsize < 0) {
    dprintf("[snfs_ping] malformed arguments.\n");
    return STAT_ERROR;
  }

  snfs_msg_req_t req;
  snfs_msg_res_t res;
  
  memset(&req, 0, sizeof(req));
  memset(&res, 0, sizeof(res));
  
  // format request
  req.type = REQ_PING;
  strncpy(req.body.ping.msg, inmsg, insize - 1);
  
  int status = remote_call(&req, sizeof(req.type) + sizeof(req.body.ping), 
			   &res, sizeof(res));
  
  // format response
  if (status < 0 || res.status != RES_OK) {
    return STAT_ERROR;
  }

  strncpy(outmsg, res.body.ping.msg, outsize);
  
  return STAT_OK;
}


snfs_call_status_t snfs_lookup(snfs_fhandle_t dir, char* name, 
			       snfs_fhandle_t* file, unsigned* fsize)
{
  if (name == NULL || file == NULL || fsize == NULL) {
    dprintf("[snfs_lookup] malformed arguments.\n");
    return STAT_ERROR;
  }

  snfs_msg_req_t req;
  snfs_msg_res_t res;

  memset(&req, 0, sizeof(req));
  memset(&res, 0, sizeof(res));

  //format request
  req.type = REQ_LOOKUP;
  req.body.lookup.dir = dir;
  strncpy(req.body.lookup.fname, name, MAX_FILE_NAME_SIZE - 1);

  int status = remote_call(&req, sizeof(req.type) + sizeof(req.body.lookup), 
			   &res, sizeof(res));

  //format response
  if (status < 0 || res.status != RES_OK) {
    return STAT_ERROR;
  }

  *file = res.body.lookup.file;
  *fsize = res.body.lookup.fsize;

  return STAT_OK;
}


snfs_call_status_t snfs_read(snfs_fhandle_t fhandle, unsigned offset,
			     unsigned count, char* buffer, int* nread)
{
  if (buffer == NULL || nread == NULL) {
    dprintf("[snfs_read] malformed arguments.\n");
    return STAT_ERROR;
  }

  snfs_msg_req_t req;
  snfs_msg_res_t res;

  memset(&req, 0, sizeof(req));
  memset(&res, 0, sizeof(res));

  //format request
  req.type = REQ_READ;
  req.body.read.fhandle = fhandle;
  req.body.read.offset = offset;
  req.body.read.count = count;

  int status = remote_call(&req, sizeof(req.type) + sizeof(req.body.read), 
			   &res, sizeof(res));

  //format response
  if (status < 0 || res.status != RES_OK) {
    return STAT_ERROR;
  }

  memcpy(buffer, res.body.read.data, res.body.read.nread);
  *nread = res.body.read.nread;

  return STAT_OK;
}


snfs_call_status_t snfs_write(snfs_fhandle_t fhandle, unsigned offset, 
			      unsigned count, char* buffer,
			      unsigned int* fsize)
{
  if (buffer == NULL || fsize == NULL) {
    dprintf("[snfs_write] malformed arguments.\n");
    return STAT_ERROR;
  }

  snfs_msg_req_t req;
  snfs_msg_res_t res;

  memset(&req, 0, sizeof(req));
  memset(&res, 0, sizeof(res));

  //format request
  req.type = REQ_WRITE;
  req.body.write.fhandle = fhandle;
  req.body.write.offset = offset;
  req.body.write.count = count;
  memcpy(req.body.write.data, buffer, count);

  int status = remote_call(&req, sizeof(req.type) + sizeof(req.body.write), 
			   &res, sizeof(res));

  //format response
  if (status < 0 || res.status != RES_OK) {
    return STAT_ERROR;
  }

  *fsize = res.body.write.fsize;

  return STAT_OK;
}


snfs_call_status_t snfs_create(snfs_fhandle_t dir, char* name, 
			       snfs_fhandle_t* file)
{
  if (name == NULL || file == NULL) {
    dprintf("[snfs_create] malformed arguments.\n");
    return STAT_ERROR;
  }

  snfs_msg_req_t req;
  snfs_msg_res_t res;

  memset(&req, 0, sizeof(req));
  memset(&res, 0, sizeof(res));

  //format request
  req.type = REQ_CREATE;
  req.body.create.dir = dir;
  memcpy(req.body.create.name, name, MAX_FILE_NAME_SIZE - 1);

  int status = remote_call(&req, sizeof(req.type) + sizeof(req.body.create),
			   &res, sizeof(res));

  //format response
  if (status < 0 || res.status != RES_OK) {
    return STAT_ERROR;
  }

  *file = res.body.create.file;

  return STAT_OK;
}


snfs_call_status_t snfs_mkdir(snfs_fhandle_t dir, char* name, 
			      snfs_fhandle_t* file)
{
  if (name == NULL || file == NULL) {
    dprintf("[snfs_mkdir] malformed arguments.\n");
    return STAT_ERROR;
  }

  snfs_msg_req_t req;
  snfs_msg_res_t res;

  memset(&req, 0, sizeof(req));
  memset(&res, 0, sizeof(res));

  //format request
  req.type = REQ_MKDIR;
  req.body.mkdir.dir = dir;
  strncpy(req.body.mkdir.file, name, MAX_FILE_NAME_SIZE);

  int status = remote_call(&req, sizeof(req.type) + sizeof(req.body.mkdir),
			   &res, sizeof(res));

  //format response
  if (status < 0 || res.status != RES_OK) {
    return STAT_ERROR;
  }

  *file = res.body.mkdir.newdirid;

  return STAT_OK;
}


snfs_call_status_t snfs_readdir(snfs_fhandle_t dir, unsigned cmax, 
				snfs_dir_entry_t* list, unsigned* count)
{
  if (list == NULL || count == NULL) {
    dprintf("[snfs_readdir] malformed arguments.\n");
    return STAT_ERROR;
  }

  snfs_msg_req_t req;
  snfs_msg_res_t res;

  memset(&req, 0, sizeof(req));
  memset(&res, 0, sizeof(res));

  //format request
  req.type = REQ_READDIR;
  req.body.readdir.dir = dir;
  req.body.readdir.cmax = cmax;

  int status = remote_call(&req, sizeof(req.type) + sizeof(req.body.readdir),
			   &res, sizeof(res));

  //format response
  if (status < 0 || res.status != RES_OK) {
    return STAT_ERROR;
  }

  memcpy(list, res.body.readdir.list,
	 sizeof(snfs_dir_entry_t) * res.body.readdir.count);
  *count = res.body.readdir.count;

  return STAT_OK;
}


void snfs_finish()
{
   close(Cli_sock);
   unlink(Cli_addr.sun_path);
}
