/* 
 * SNFS Server
 * 
 * snfs_server.c
 *
 * Server implementing SNFS services. Current implementation is
 * single-threaded.
 * 
 * Read the project specification for futher information.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sthread.h>

// SNFS includes
#include <snfs_proto.h>
#include "snfs.h"


#ifndef SERVER_SOCK
#define SERVER_SOCK "/tmp/server.socket"
#endif


/*
 * SNFS Services
 *
 * all services are registered in a global table which is
 * used during request resolution to find the handler that
 * serves the request
 *
 * service handlers are implemented in snfs.c
 */


// number of handling threads
#define NUM_TC 5

// number of slots at the buffer
#define RING_SIZE 10

// number of request handlers
#define NUM_REQ_HANDLERS 7


struct {
  snfs_msg_type_t type;
  snfs_handler_t handler;
} Service[NUM_REQ_HANDLERS] = {
  {REQ_PING, snfs_ping},
  {REQ_LOOKUP, snfs_lookup},
  {REQ_READ, snfs_read},
  {REQ_WRITE, snfs_write},
  {REQ_CREATE, snfs_create},
  {REQ_MKDIR, snfs_mkdir},
  {REQ_READDIR, snfs_readdir}
};


// type of a server request
typedef struct {
  int sockfd, reqsz;
  struct sockaddr_un cliaddr;
  socklen_t clilen;
  snfs_msg_req_t req;
} srv_req_t;


/*
 * Internal private global variables
 */

// requests buffer
static srv_req_t Request[RING_SIZE];

// controls access to the buffer
static sthread_sem_t Mutex;

// counts empty buffer slots
static sthread_sem_t Sem_empty;

// counts full buffer slots
static sthread_sem_t Sem_full;

// controls access to the counter
static sthread_sem_t Mutex_num;

// number of requests to process and in process
static int num_requests;

// socket state (1 = blocking, 0 = non blocking)
static int socket_blk_state;


/*
 * Internal auxiliary functions for server request management
 */


static void srv_req_insert(srv_req_t* srv_req)
{
  static int next = 0;
  memcpy(&Request[next], srv_req, sizeof(srv_req_t));
  ++next;
  next %= RING_SIZE;
}


static void srv_req_remove(srv_req_t* srv_req)
{
  static int next = 0;
  memcpy(srv_req, &Request[next], sizeof(srv_req_t));
  ++next;
  next %= RING_SIZE;
}


/*
 * Internal auxiliary funtions for communication management
 */


void srv_init_socket(int* sockfd, struct sockaddr_un* servaddr)
{
  // creates socket datagram domain unix
  if ((*sockfd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0){
    printf("[snfs_srv] socket error: %s.\n", strerror(errno));
    exit(-1);
  }

  // structure address cleaning
  bzero(servaddr, sizeof(*servaddr));
  servaddr->sun_family = AF_UNIX;
  strcpy(servaddr->sun_path, SERVER_SOCK);
  
  // if exists, deletes socket file name
  if (unlink(servaddr->sun_path) < 0 && errno != ENOENT) {
    printf("[snfs_srv] unlink error: %s.\n", strerror(errno));
    exit(-1);
  }

  // binds socket to address
  if (bind(*sockfd, (struct sockaddr *) servaddr, sizeof(*servaddr)) < 0){
    printf("[snfs_srv] unbind error: %s.\n", strerror(errno));
    exit(-1);
  }
}


int srv_recv_request(int sockfd, snfs_msg_req_t* req, 
		     struct sockaddr_un* cliaddr, socklen_t* clilen)
{
  *clilen = sizeof(*cliaddr);
  int reqsz = recvfrom(sockfd, (void*)req, sizeof(*req), 0, 
		       (struct sockaddr *)cliaddr, clilen);
  if (reqsz == 0) {
    printf("[snfs_srv] request error.\n");
    return 0;
  }
  if (reqsz == -1) {
    // no message available to read
    return 0;
  }
  if (reqsz < 0) {
    printf("[snfs_srv] recvfrom error: %s.\n", strerror(errno));
    exit(-1);
  }
  return reqsz;
}


void srv_send_response(int sockfd, snfs_msg_res_t* res, int ressz, 
		       struct sockaddr_un* cliaddr, socklen_t clilen)
{
  int status;
  do {
    // try to send the response
    status = sendto(sockfd, res, ressz, 0, (struct sockaddr *)cliaddr, clilen);
  } while (status == -1 && errno == EAGAIN);
  if (status < 0) {
    printf("[snfs_srv] sendto error: %s.\n", strerror(errno));
    exit(-1);
  }
  if (status != ressz) {
    printf("[snfs_srv] message size mismatch.\n");
  }
}


void* srv_handler_request(void * arg)
{
  int ressz, req_i;
  srv_req_t srv_req;
  snfs_msg_res_t res;

  while (1) {
    sthread_sem_wait(Sem_full);
    sthread_sem_wait(Mutex);

    // remove request from buffer
    srv_req_remove(&srv_req);

    sthread_sem_post(Mutex);
    sthread_sem_post(Sem_empty);

    // find request handler
    req_i = -1;
    for (int i = 0; i < NUM_REQ_HANDLERS; i++) {
      if (srv_req.req.type == Service[i].type) {
	req_i = i;
	break;
      }
    }
    
    // serve the request
    if (req_i == -1) {
      res.status = RES_UNKNOWN;
      ressz = sizeof(res) - sizeof(res.body);
      printf("[snfs_srv] unknown request.\n");
    } else {
      Service[req_i].handler(&srv_req.req, srv_req.reqsz, &res, &ressz);
    }
    
    // send response to client
    srv_send_response(srv_req.sockfd, &res, ressz, &srv_req.cliaddr,
		      srv_req.clilen);
  }

  sthread_sem_wait(Mutex_num);
  num_requests--;
  sthread_sem_post(Mutex_num);
  
  return NULL;
}


/*
 * SNFS server main
 */

int main(int argc, char **argv)
{
  struct sockaddr_un servaddr;
  srv_req_t srv_req;

  // initialize control structures
  sthread_init();
  Mutex = sthread_sem_init(1); 
  Sem_empty = sthread_sem_init(RING_SIZE);
  Sem_full = sthread_sem_init(0);
  Mutex_num = sthread_sem_init(1);
  num_requests = 0;
  socket_blk_state = 1;

  // initialize the request handling threads
  for (int i = 0; i < NUM_TC; i++) {
    if (sthread_create(srv_handler_request, (void *)NULL) == 0) {
      printf("[snfs_srv] thread creating error.\n");
      exit(-1);
    }
  }

  // initialize SNFS layer
  snfs_init(argc,argv);

  // initialize communications
  srv_init_socket(&srv_req.sockfd, &servaddr);

  int flags = fcntl(srv_req.sockfd, F_GETFL);

  while (1) {
    sthread_sem_wait(Mutex_num);
    
    // tests if there are no requests, and if so, set socket as blocking
    if (socket_blk_state == 0 && num_requests == 0) {
      fcntl(srv_req.sockfd, F_SETFL, flags & ~O_NONBLOCK);
      socket_blk_state = 1;
    }

    sthread_sem_post(Mutex_num);

    // read request from client
    if ((srv_req.reqsz = srv_recv_request(srv_req.sockfd, &srv_req.req,
					  &srv_req.cliaddr,
					  &srv_req.clilen)) == 0) {
      sthread_yield();
      continue;
    }

    if (socket_blk_state == 1) {
      // set socket as non blocking
      fcntl(srv_req.sockfd, F_SETFL, flags | O_NONBLOCK);
    }

    sthread_sem_wait(Sem_empty);
    sthread_sem_wait(Mutex);

    // insert request in buffer
    srv_req_insert(&srv_req);

    sthread_sem_post(Mutex);

    sthread_sem_wait(Mutex_num);
    num_requests++;
    sthread_sem_post(Mutex_num);

    sthread_sem_post(Sem_full);
  }
}
