/*
 * sthread.c - Implements the public API (the functions defined in
 *             include/sthread.h). Since sthreads supports two implementations
 *             (pthreads and student supplied user-level threads), this
 *             just consists of dispatching the calls to the implementation
 *             that the application choose.
 *
 */

#include <config.h>

#include <assert.h>

#include <sthread.h>
#include <sthread_pthread.h>
#include <sthread_user.h>

#ifdef USE_PTHREADS
#define IMPL_CHOOSE(pthread, user) pthread
#else
#define IMPL_CHOOSE(pthread, user) user
#endif

void sthread_init(void) {
  IMPL_CHOOSE(sthread_pthread_init(), sthread_user_init());
}

sthread_t sthread_create(sthread_start_func_t start_routine, void *arg) {
  sthread_t newth;
  IMPL_CHOOSE(newth = sthread_pthread_create(start_routine, arg),
	      newth = sthread_user_create(start_routine, arg));
  return newth;
}

void sthread_exit(void *ret) {
  IMPL_CHOOSE(sthread_pthread_exit(ret), sthread_user_exit(ret));
}

void sthread_yield(void) {
  IMPL_CHOOSE(sthread_pthread_yield(), sthread_user_yield());
}

int sthread_sleep(int time) {
   return IMPL_CHOOSE(sthread_pthread_sleep(time),sthread_user_sleep(time));
}

int sthread_join(sthread_t thread, void **value_ptr) {
  return IMPL_CHOOSE(sthread_pthread_join(thread,value_ptr),sthread_user_join(thread,value_ptr));
}

/**********************************************************************/
/* Synchronization Primitives: Mutexs and Condition Variables         */
/**********************************************************************/


sthread_mutex_t sthread_mutex_init() {
  sthread_mutex_t lock;
  IMPL_CHOOSE(lock = sthread_pthread_mutex_init(),
	      lock = sthread_user_mutex_init());
  return lock;
}

void sthread_mutex_free(sthread_mutex_t lock) {
  IMPL_CHOOSE(sthread_pthread_mutex_free(lock),
	      sthread_user_mutex_free(lock));
}

void sthread_mutex_lock(sthread_mutex_t lock) {
  IMPL_CHOOSE(sthread_pthread_mutex_lock(lock),
	      sthread_user_mutex_lock(lock));
}

void sthread_mutex_unlock(sthread_mutex_t lock) {
  IMPL_CHOOSE(sthread_pthread_mutex_unlock(lock),
	      sthread_user_mutex_unlock(lock));
}

sthread_sem_t sthread_sem_init(int value) {
  sthread_sem_t sem;
  IMPL_CHOOSE(sem = sthread_pthread_sem_init(value),
	      sem = sthread_user_sem_init(value));
  return sem;
}

void sthread_sem_free(sthread_sem_t sem) {
  IMPL_CHOOSE(sthread_pthread_sem_free(sem),
	      sthread_user_sem_free(sem));
}

void sthread_sem_wait(sthread_sem_t sem) {
  IMPL_CHOOSE(sthread_pthread_sem_wait(sem),
	      sthread_user_sem_wait(sem));
}

void sthread_sem_post(sthread_sem_t sem) {
  IMPL_CHOOSE(sthread_pthread_sem_post(sem),
	      sthread_user_sem_post(sem));
}


