/* Simplethreads Instructional Thread Package
 * 
 * sthread_user.c
 *
 * Implements the sthread API using user-level threads.
 *
 */


#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>

#include <sthread.h>
#include <sthread_user.h>
#include <sthread_ctx.h>
#include <sthread_time_slice.h>
#include <sthread_user.h>
#include "queue.h"
#include "pqueue.h"


#define CLOCK_TICK 10000

extern int good_interrupts;
extern int dropped_interrupts;


// thread states
typedef enum {
  RUNNING = 1,
  ALIVE = 2,
  SLEEP = 3,
  WAIT = 4,
  BLOCKED = 5,
  ZOMBIE = 6,
  DEAD = 7} sthread_state_t;


struct _sthread {
  int tid;
  void* args;
  sthread_start_func_t start_routine_ptr;
  sthread_ctx_t *saved_ctx;
  queue_t *wait_thr_list;
  void** ret;
  sthread_state_t state;
};


static int Tid_gen;                   // TID generator
static struct _sthread *Active_thr;   // active thread
static queue_t *Exe_thr_list;         // executable thread queue
static struct _sthread *Dead_thr;     // dead thread
static int Lib_initted = 0;           // Flag to test if library was initted
static pqueue_t *Sleep_thr_list;      // sleeping thread priority queue
static queue_t *Dead_thr_list;         // dead/zombie thread queue


/* --------------------------------------------------------------------------*
 * Creating and Scheduling Threads                                           *
 * ------------------------------------------------------------------------- */


void sthread_user_init(void)
{
  // initialize global state
  Tid_gen = 1;
  Exe_thr_list = queue_create();
  Dead_thr = NULL; 
  Sleep_thr_list = pqueue_create(1024);
  Dead_thr_list = queue_create();

  // capture the main thread context
  struct _sthread *main_thread = malloc(sizeof(struct _sthread));
  main_thread->tid = Tid_gen++;
  main_thread->args = NULL;
  main_thread->start_routine_ptr = NULL;
  main_thread->saved_ctx = sthread_new_blank_ctx();
  main_thread->ret = NULL;
  main_thread->state = RUNNING;
   
  // set the main thread as the active thread
  Active_thr = main_thread;

  Lib_initted = 1;

  // initialize time slices
  sthread_time_slices_init(sthread_user_yield, CLOCK_TICK);
}


void sthread_aux_start(void)
{
  splx(LOW);
  Active_thr->ret = Active_thr->start_routine_ptr(Active_thr->args);
  splx(HIGH);
  sthread_user_exit(Active_thr->ret);
}


sthread_t sthread_user_create(sthread_start_func_t start_routine, void *arg)
{
  if (!Lib_initted) {
    printf("[sthread_user_create] Library is not initialized.\n");
    exit(-1);
  }

  if (start_routine == NULL) {
    printf("[sthread_user_create] Start routine is null.\n");
    return NULL;
  }
   
  // create context for the new thread
  struct _sthread *new_thread = 
    (struct _sthread*)malloc(sizeof(struct _sthread));
  new_thread->args = arg;
  new_thread->start_routine_ptr = start_routine;
  new_thread->saved_ctx = sthread_new_ctx(sthread_aux_start);
  new_thread->wait_thr_list = queue_create();
  new_thread->ret = NULL;
  new_thread->state = ALIVE;
  
  // assign a TID to the thread and make it executable
  splx(HIGH);
  new_thread->tid = Tid_gen++;
  queue_enqueue(Exe_thr_list, new_thread);
  splx(LOW);

  return new_thread;
}


void sthread_user_yield(void)
{
  if (!Lib_initted) {
    printf("[sthread_user_yield] Library is not initialized.\n");
    exit(-1);
  }
   
  splx(HIGH);
  int current_time = good_interrupts + dropped_interrupts;

  while (!pqueue_is_empty(Sleep_thr_list)
	&& pqueue_minimum(Sleep_thr_list) <= current_time) {
    queue_enqueue(Exe_thr_list, pqueue_dequeue(Sleep_thr_list));
  }

  if (!queue_is_empty(Exe_thr_list)) {
    struct _sthread *old_thr = Active_thr;
    Active_thr->state = ALIVE;
    queue_enqueue(Exe_thr_list, Active_thr);
    Active_thr = queue_dequeue(Exe_thr_list);
    Active_thr->state = RUNNING;
    sthread_switch(old_thr->saved_ctx, Active_thr->saved_ctx);
  }
  splx(LOW);
}


void sthread_user_exit(void *ret)
{
  if (!Lib_initted) {
    printf("[sthread_user_exit] Library is not initialized.\n");
    exit(-1);
  }

  splx(HIGH);
  if (Dead_thr != NULL) {
    // free the structures of the last dead thread
    sthread_free_ctx(Dead_thr->saved_ctx);
    queue_destroy(Dead_thr->wait_thr_list);
    Dead_thr = NULL;
  }

  // mark this thread to be destroyed
  Dead_thr = Active_thr;

  // store the thread structure at the dead/zombie thread list
  queue_enqueue(Dead_thr_list, Active_thr);

  if (!queue_is_empty(Active_thr->wait_thr_list)) {
    while (!queue_is_empty(Active_thr->wait_thr_list)) {
      struct _sthread *wait_thr = queue_dequeue(Active_thr->wait_thr_list);

      if (Active_thr->state != DEAD) {
        Active_thr->state = DEAD;

        if (wait_thr->ret != NULL) {
          *(wait_thr->ret) = ret;
        }
      }
      queue_enqueue(Exe_thr_list, wait_thr);
      wait_thr->state = ALIVE;
    }
  } else {
    // the thread will became a zombie thread
    Active_thr->ret = ret;
    Active_thr->state = ZOMBIE;
  }

  struct _sthread *old_thr = Active_thr;
  if (queue_is_empty(Exe_thr_list)) {
    if (pqueue_is_empty(Sleep_thr_list)) {
      // the last thread terminated, exit the process
      queue_destroy(Exe_thr_list);
      pqueue_destroy(Sleep_thr_list);
      while (!queue_is_empty(Dead_thr_list)) {
        // free the dead/zombie threads
        free(queue_dequeue(Dead_thr_list));
      }
      queue_destroy(Dead_thr_list);
      exit(0);
    } else {
      int current_time = good_interrupts + dropped_interrupts;
      int sleep_time = pqueue_minimum(Sleep_thr_list) - current_time;
      dropped_interrupts += sleep_time;
      usleep(sleep_time * CLOCK_TICK);
	 
      // switch to the next executable thread
      Active_thr = pqueue_dequeue(Sleep_thr_list);
      Active_thr->state = RUNNING;
      sthread_switch(old_thr->saved_ctx, Active_thr->saved_ctx);
    }
  } else {
    // switch to the next executable thread
    Active_thr = queue_dequeue(Exe_thr_list);
    Active_thr->state = RUNNING;
    sthread_switch(old_thr->saved_ctx, Active_thr->saved_ctx);
  }
  splx(LOW);
}



/* --------------------------------------------------------------------------*
 * Sleep and Join Primitives                                                 *
 * ------------------------------------------------------------------------- */


int sthread_user_sleep(int time)
{
  if (time > 0) {
    splx(HIGH);
    int current_time = good_interrupts + dropped_interrupts;
    struct _sthread *old_thr = Active_thr;
    Active_thr->state = SLEEP;
    pqueue_enqueue(Sleep_thr_list, Active_thr,
		   current_time + 1000 * time / CLOCK_TICK);
 
    if (!queue_is_empty(Exe_thr_list)) {
      Active_thr = queue_dequeue(Exe_thr_list);
    } else {
      int sleep_time = pqueue_minimum(Sleep_thr_list) - current_time;

      dropped_interrupts += sleep_time;
      usleep(sleep_time * CLOCK_TICK);

      Active_thr = pqueue_dequeue(Sleep_thr_list);
    }

    Active_thr->state = RUNNING;
    sthread_switch(old_thr->saved_ctx, Active_thr->saved_ctx);
    splx(LOW);
    return 0;
  } else if (time == 0) {
    return 0;
  } else {
    printf("[sthread_user_sleep] Argument was less than zero.\n");
    return -1;
  }
}


int sthread_user_join(sthread_t thread, void **value_ptr)
{
  splx(HIGH);
  if (thread == NULL || thread == Active_thr || thread->state == DEAD) {
    splx(LOW);
    return -1;
  }

  // zombie thread, thread to join has already exit
  if (thread->state == ZOMBIE) {
    thread->state = DEAD;

    if (value_ptr != NULL) {
      *value_ptr = thread->ret;
    }

    splx(LOW);
    return 0;
  }

  if (queue_is_empty(Exe_thr_list) && pqueue_is_empty(Sleep_thr_list)) {
    splx(LOW);
    return -1;
  }

  Active_thr->ret = value_ptr;
  Active_thr->state = WAIT;
  struct _sthread *old_thr = Active_thr;
  queue_enqueue(thread->wait_thr_list, Active_thr);
 
  if (!queue_is_empty(Exe_thr_list)) {
    Active_thr = queue_dequeue(Exe_thr_list);
  } else {
    int current_time = good_interrupts + dropped_interrupts;
    int sleep_time = pqueue_minimum(Sleep_thr_list) - current_time;
     
    dropped_interrupts += sleep_time;
    usleep(sleep_time * CLOCK_TICK);
     
    Active_thr = pqueue_dequeue(Sleep_thr_list);
  }

  Active_thr->state = RUNNING;
  sthread_switch(old_thr->saved_ctx, Active_thr->saved_ctx);

  splx(LOW);
  return 0;
}


/* --------------------------------------------------------------------------*
 * Synchronization Primitives                                                *
 * ------------------------------------------------------------------------- */


/*
 * Semaphore implementation
 */


struct _sthread_sem {
  int value;
  queue_t *thr_list;
};


sthread_sem_t sthread_user_sem_init(int value)
{
  sthread_sem_t new_sem;

  new_sem = (sthread_sem_t)malloc(sizeof(struct _sthread_sem));
  new_sem->value = value;
  new_sem->thr_list = queue_create();

  return new_sem;
}


void sthread_user_sem_free(sthread_sem_t sem)
{
  if (!queue_is_empty(sem->thr_list)) {
    printf("[sthread_user_sem_free] Thread list is not empty.\n");
    exit(-1);
  }

  queue_destroy(sem->thr_list);
  free(sem);
}


void sthread_user_sem_wait(sthread_sem_t sem)
{
  splx(HIGH);
  if (sem->value > 0) {
    sem->value--;
  } else {
      struct _sthread *old_thr = Active_thr;
      Active_thr->state = BLOCKED;
      queue_enqueue(sem->thr_list, Active_thr);

      if (!queue_is_empty(Exe_thr_list)) {
        Active_thr = queue_dequeue(Exe_thr_list);
      } else {
        int current_time = good_interrupts + dropped_interrupts;
        int sleep_time = pqueue_minimum(Sleep_thr_list) - current_time;
        
        dropped_interrupts += sleep_time;
        usleep(sleep_time * CLOCK_TICK);
        
        Active_thr = pqueue_dequeue(Sleep_thr_list);
      }

      Active_thr->state = RUNNING;
      sthread_switch(old_thr->saved_ctx, Active_thr->saved_ctx);
  }
  splx(LOW);
}


void sthread_user_sem_post(sthread_sem_t sem)
{
  splx(HIGH);
  if (!queue_is_empty(sem->thr_list)) {
    struct _sthread *old_thr = queue_dequeue(sem->thr_list);
    old_thr->state = ALIVE;
    queue_enqueue(Exe_thr_list, old_thr);
  }
  else {
    sem->value++;
  }
  splx(LOW);
}



/*
 * Mutex implementation
 */


struct _sthread_mutex {
  /* DO NOT IMPLEMENT */
};


sthread_mutex_t sthread_user_mutex_init()
{
  /* DO NOT IMPLEMENT */
  return NULL;
}


void sthread_user_mutex_free(sthread_mutex_t lock)
{
  /* DO NOT IMPLEMENT */
}


void sthread_user_mutex_lock(sthread_mutex_t lock)
{
  /* DO NOT IMPLEMENT */
}


void sthread_user_mutex_unlock(sthread_mutex_t lock)
{
  /* DO NOT IMPLEMENT */
}
